# -*- coding: utf-8 -*-
from plone.app.caching.testing import getToken
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.caching.interfaces import ICacheSettings
from plone.registry.interfaces import IRegistry
from plone.testing.zope import Browser
from zope.component import getUtility
from zope.globalrequest import setRequest

import unittest


class TestOperationDefault(unittest.TestCase):
    """
    Test various edge cases.
    """

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        setRequest(self.portal.REQUEST)

        self.registry = getUtility(IRegistry)
        self.cacheSettings = self.registry.forInterface(ICacheSettings)
        self.cacheSettings.enabled = True

    def tearDown(self):
        setRequest(None)

    def test_last_modified_no_etags(self):
        """
        When a new content type is added, the resulting page should not be
        cached since it has messages. However, it should only trigger an etag
        if its been configured to use etags
        """
        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ('Manager',))
        self.portal.invokeFactory('Folder', 'f1')
        self.portal['f1'].title = u'Folder one'
        self.portal['f1'].description = u'Folder one description'
        self.portal['f1'].reindexObject()

        self.cacheSettings.operationMapping = {
            'plone.content.itemView': 'plone.app.caching.weakCaching'}
        self.registry['plone.app.caching.weakCaching.lastModified'] = True
        self.registry['plone.app.caching.weakCaching.etags'] = None

        import transaction
        transaction.commit()

        # log in and create a content type
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open('{0}/++add++Document'.format(
            self.portal['f1'].absolute_url()),)
        browser.getControl(
            name='form.widgets.IDublinCore.title').value = 'dummy content'
        browser.getControl('Save').click()
        self.assertNotIn('Etag', browser.headers)

        # now set up etags and make sure that a header is added
        self.registry['plone.app.caching.weakCaching.etags'] = (
            'lastModified',
        )
        import transaction
        transaction.commit()
        browser.open('{0}/dummy-content/edit?_authenticator={1}'.format(
            self.portal['f1'].absolute_url(),
            getToken(TEST_USER_NAME)))
        browser.getControl(
            name='form.widgets.IDublinCore.title').value = 'dummy content'
        browser.getControl('Save').click()
        self.assertIn('Etag', browser.headers)



# -*- coding: utf-8 -*-
from plone.app.caching.interfaces import _
from plone.app.caching.operations.utils import cachedResponse
from plone.app.caching.operations.utils import cacheInRAM
from plone.app.caching.operations.utils import cacheStop
from plone.app.caching.operations.utils import doNotCache
from plone.app.caching.operations.utils import fetchFromRAMCache
from plone.app.caching.operations.utils import getContext
from plone.app.caching.operations.utils import getETagAnnotation
from plone.app.caching.operations.utils import getLastModifiedAnnotation
from plone.app.caching.operations.utils import isModified
from plone.app.caching.operations.utils import notModified
from plone.app.caching.operations.utils import parseDateTime
from plone.app.caching.operations.utils import setCacheHeaders
from plone.app.caching.operations.utils import visibleToRole
from plone.caching.interfaces import ICachingOperation
from plone.caching.interfaces import ICachingOperationType
from plone.caching.utils import lookupOptions
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import provider
from zope.publisher.interfaces.http import IHTTPRequest

import datetime
import random
import time


@implementer(ICachingOperation)
@provider(ICachingOperationType)
@adapter(Interface, Interface)
class BaseCaching(object):
    """A generic caching operation class that can do pretty much all the usual
    caching operations based on options settings. For UI simplicity, it might
    be easier to subclass this in your custom operations to set a few default
    operations.

    Generic options (Default value for each is None):

    ``maxage``
        is the maximum age of the cached item, in seconds..

    ``smaxage``
        is the maximum age of the cached item in proxies, in seconds.

    ``etags''
        is a list of etag components to use when constructing an etag.

    ``lastModified``
        is a boolean indicating whether to set a Last-Modified header
        and turn on 304 responses.

    ``ramCache``
        is a boolean indicating whether to turn on RAM caching for this item.
        Etags are only required if the URL is not specific enough to ensure
        uniqueness.

    ``vary``
        is a string to add as a Vary header value in the response.
    """

    title = _(u'Generic caching')
    description = _(
        u'Through this operation, all standard caching functions '
        u'can be performed via various combinations of the optional '
        u'parameter settings. For most cases, it\'s probably easier '
        u'to use one of the other simpler operations (Strong caching, '
        u'Moderate caching, Weak caching, or No caching).'
    )
    prefix = 'plone.app.caching.baseCaching'
    options = ('maxage', 'smaxage', 'etags', 'lastModified',
               'ramCache', 'vary', 'anonOnly')

    # Default option values
    maxage = smaxage = etags = vary = None
    lastModified = ramCache = anonOnly = False

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def interceptResponse(self, rulename, response, class_=None):
        options = lookupOptions(class_ or self.__class__, rulename)

        etags = options.get('etags') or self.etags
        anonOnly = options.get('anonOnly', self.anonOnly)
        ramCache = options.get('ramCache', self.ramCache)
        lastModified = options.get('lastModified', self.lastModified)

        # Add the ``anonymousOrRandom`` etag if we are anonymous only
        if anonOnly:
            if etags is None:
                etags = ['anonymousOrRandom']
            elif 'anonymousOrRandom' not in etags:
                etags = tuple(etags) + ('anonymousOrRandom',)

        etag = getETagAnnotation(self.published, self.request, keys=etags)
        lastModified = getLastModifiedAnnotation(
            self.published, self.request, lastModified=lastModified)

        # Remove Range from request if the If-Range condition is not fulfilled
        if_range = self.request.environ.get('HTTP_IF_RANGE', '').strip('"')
        if if_range:
            if 'HTTP_RANGE' in self.request.environ:
                if_range_dt = parseDateTime(if_range)
                delta_sec = datetime.timedelta(seconds=1)
                if if_range_dt and (lastModified - if_range_dt) < delta_sec:
                    pass
                elif if_range == etag:
                    pass
                else:
                    del self.request.environ['HTTP_RANGE']
            # If-Range check is done here so we could remove it from the request
            del self.request.environ['HTTP_IF_RANGE']

        # Check for cache stop request variables
        if cacheStop(self.request, rulename):
            return None

        # Check if this should be a 304 response
        if not isModified(self.request, etag=etag, lastModified=lastModified):
            return notModified(
                self.published,
                self.request,
                response,
                etag=etag,
                lastModified=lastModified,
            )

        # Check if this is in the ram cache
        if ramCache:
            context = getContext(self.published)
            portal_state = getMultiAdapter(
                (context, self.request), name=u'plone_portal_state')

            if portal_state.anonymous():
                cached = fetchFromRAMCache(
                    self.request, etag=etag, lastModified=lastModified)
                if cached is not None:
                    return cachedResponse(
                        self.published,
                        self.request,
                        response,
                        *cached
                    )

        return None

    def modifyResponse(self, rulename, response, class_=None):
        options = lookupOptions(class_ or self.__class__, rulename)

        maxage = options.get('maxage', self.maxage)
        smaxage = options.get('smaxage', self.smaxage)
        etags = options.get('etags') or self.etags

        anonOnly = options.get('anonOnly', self.anonOnly)
        ramCache = options.get('ramCache', self.ramCache)
        vary = options.get('vary', self.vary)

        # Add the ``anonymousOrRandom`` etag if we are anonymous only
        if anonOnly:
            if etags is None:
                etags = ['anonymousOrRandom']
            elif 'anonymousOrRandom' not in etags:
                etags = tuple(etags) + ('anonymousOrRandom',)

        etag = getETagAnnotation(self.published, self.request, etags)
        lastModified = getLastModifiedAnnotation(
            self.published, self.request, options['lastModified'])

        # Check for cache stop request variables
        if cacheStop(self.request, rulename):
            # only stop with etags if configured
            if etags:
                etag = '{0}{1}'.format(time.time(), random.randint(0, 1000))
                return setCacheHeaders(
                    self.published,
                    self.request,
                    response,
                    etag=etag,
                )
            # XXX: should there be an else here? Last modified works without
            #      extra headers.
            #      Are there other config options?

        # Do the maxage/smaxage settings allow for proxy caching?
        proxyCache = smaxage or (maxage and smaxage is None)

        # Check if the content can be cached in shared caches
        public = True
        if ramCache or proxyCache:
            if etags is not None:
                if (
                    'userid' in etags or
                    'anonymousOrRandom' in etags or
                    'roles' in etags

                ):
                    context = getContext(self.published)
                    portal_state = getMultiAdapter(
                        (context, self.request),
                        name=u'plone_portal_state',
                    )
                    public = portal_state.anonymous()
            public = public and visibleToRole(self.published, role='Anonymous')

        if proxyCache and not public:
            # This is private so keep it out of both shared and browser caches
            maxage = smaxage = 0

        setCacheHeaders(
            self.published,
            self.request,
            response,
            maxage=maxage,
            smaxage=smaxage,
            etag=etag,
            lastModified=lastModified,
            vary=vary,
        )

        if ramCache and public:
            cacheInRAM(self.published, self.request, response,
                       etag=etag, lastModified=lastModified)


@provider(ICachingOperationType)
class WeakCaching(BaseCaching):
    """Weak caching operation. A subclass of the generic BaseCaching
    operation to help make the UI approachable by mortals
    """

    title = _(u'Weak caching')
    description = _(
        u'Cache in browser but expire immediately and enable 304 '
        u'responses on subsequent requests. 304\'s require configuration '
        u'of the \'Last-modified\' and/or \'ETags\' settings. If '
        u'Last-Modified  header is insufficient to ensure freshness, turn on '
        u'ETag checking by listing each ETag components that should be used '
        u'to construct the ETag header. To also cache public responses in '
        u'Zope memory, set \'RAM cache\' to True.'
    )
    prefix = 'plone.app.caching.weakCaching'
    sort = 3

    # Configurable options
    options = ('etags', 'lastModified', 'ramCache', 'vary', 'anonOnly')

    # Default option values
    maxage = 0
    smaxage = etags = vary = None
    lastModified = ramCache = anonOnly = False


@provider(ICachingOperationType)
class ModerateCaching(BaseCaching):
    """Moderate caching operation. A subclass of the generic BaseCaching
    operation to help make the UI approachable by mortals
    """

    title = _(u'Moderate caching')
    description = _(
        u'Cache in browser but expire immediately (same as \'weak caching\'), '
        u'and cache in proxy (default: 24 hrs). '
        u'Use a purgable caching reverse proxy for best results. '
        u'Caution: If proxy cannot be purged, or cannot be configured '
        u'to remove the \'s-maxage\' token from the response, then stale '
        u'responses might be seen until the cached entry expires.')
    prefix = 'plone.app.caching.moderateCaching'
    sort = 2

    # Configurable options
    options = ('smaxage', 'etags', 'lastModified',
               'ramCache', 'vary', 'anonOnly')

    # Default option values
    maxage = 0
    smaxage = 86400
    etags = vary = None
    lastModified = ramCache = anonOnly = False


@provider(ICachingOperationType)
class StrongCaching(BaseCaching):
    """Strong caching operation. A subclass of the generic BaseCaching
    operation to help make the UI approachable by mortals
    """

    title = _(u'Strong caching')
    description = _(
        u'Cache in browser and proxy (default: 24 hrs). '
        u'Caution: Only use for stable resources '
        u'that never change without changing their URL, or resources '
        u'for which temporary staleness is not critical.'
    )
    prefix = 'plone.app.caching.strongCaching'
    sort = 1

    # Configurable options
    options = ('maxage', 'smaxage', 'etags', 'lastModified',
               'ramCache', 'vary', 'anonOnly')

    # Default option values
    maxage = 86400
    smaxage = etags = vary = None
    lastModified = ramCache = anonOnly = False


@implementer(ICachingOperation)
@provider(ICachingOperationType)
@adapter(Interface, IHTTPRequest)
class NoCaching(object):
    """A caching operation that tries to keep the response
    out of all caches.
    """

    title = _(u'No caching')
    description = _(u'Use this operation to keep the response '
                    u'out of all caches.')
    prefix = 'plone.app.caching.noCaching'
    sort = 4
    options = ()

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def interceptResponse(self, rulename, response):
        return None

    def modifyResponse(self, rulename, response):
        doNotCache(self.published, self.request, response)



# -*- coding: utf-8 -*-
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.textfield.value import RichTextValue
from plone.caching.interfaces import ICacheSettings
from plone.registry.interfaces import IRegistry
from plone.testing.zope import Browser
from zope.component import getUtility
from zope.globalrequest import setRequest

import transaction
import unittest


class TestOperationParameters(unittest.TestCase):
    """This test aims to test the effect of changing various caching operation
    parameters.
    """

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        setRequest(self.portal.REQUEST)

        self.registry = getUtility(IRegistry)
        self.cacheSettings = self.registry.forInterface(ICacheSettings)
        self.cacheSettings.enabled = True

    def tearDown(self):
        setRequest(None)

    def test_anon_only(self):

        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ('Manager',))
        self.portal.invokeFactory('Folder', 'f1')
        self.portal['f1'].title = u'Folder one'
        self.portal['f1'].description = u'Folder one description'
        self.portal['f1'].reindexObject()

        # Add page content
        self.portal['f1'].invokeFactory('Document', 'd1')
        self.portal['f1']['d1'].title = u'Document one'
        self.portal['f1']['d1'].description = u'Document one description'
        testText = 'Testing... body one'
        self.portal['f1']['d1'].text = RichTextValue(
            testText,
            'text/plain',
            'text/html',
        )
        self.portal['f1']['d1'].reindexObject()

        # Publish the folder and page
        self.portal.portal_workflow.doActionFor(self.portal['f1'], 'publish')
        self.portal.portal_workflow.doActionFor(
            self.portal['f1']['d1'], 'publish')

        # Set pages to have weak caching and test anonymous

        self.cacheSettings.operationMapping = {
            'plone.content.itemView': 'plone.app.caching.weakCaching'}
        transaction.commit()

        # View the page as anonymous
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertIn(testText, browser.contents)
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])

        # Set pages to have moderate caching so that we can see the difference
        # between logged in and anonymous

        self.cacheSettings.operationMapping = {
            'plone.content.itemView': 'plone.app.caching.moderateCaching'}
        self.registry['plone.app.caching.moderateCaching.smaxage'] = 60
        self.registry['plone.app.caching.moderateCaching.vary'] = 'X-Anonymous'
        self.registry['plone.app.caching.moderateCaching.anonOnly'] = True

        transaction.commit()

        # View the page as anonymous
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.moderateCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertIn(testText, browser.contents)
        self.assertEqual('max-age=0, s-maxage=60, must-revalidate',
                         browser.headers['Cache-Control'])
        self.assertEqual('X-Anonymous', browser.headers['Vary'])
        self.assertNotIn('Etag', browser.headers)

        # View the page as logged-in
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.moderateCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertIn(testText, browser.contents)
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        self.assertIn('Etag', browser.headers)

        # Set pages to have strong caching so that we can see the difference
        # between logged in and anonymous

        self.cacheSettings.operationMapping = {
            'plone.content.itemView': 'plone.app.caching.strongCaching'}
        self.registry['plone.app.caching.strongCaching.vary'] = 'X-Anonymous'
        self.registry['plone.app.caching.strongCaching.anonOnly'] = True
        transaction.commit()

        # View the page as anonymous
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.strongCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertIn(testText, browser.contents)
        self.assertEqual('max-age=86400, proxy-revalidate, public',
                         browser.headers['Cache-Control'])
        self.assertEqual('X-Anonymous', browser.headers['Vary'])
        self.assertNotIn('Etag', browser.headers)

        # View the page as logged-in
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.strongCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertIn(testText, browser.contents)
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        self.assertIn('Etag', browser.headers)

        # Check an edge case that has had a problem in the past:
        # setting strongCaching maxage to zero.

        self.registry['plone.app.caching.strongCaching.maxage'] = 0
        self.registry['plone.app.caching.strongCaching.smaxage'] = 60
        transaction.commit()

        # View the page as anonymous
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('max-age=0, s-maxage=60, must-revalidate',
                         browser.headers['Cache-Control'])



# -*- coding: utf-8 -*-
from dateutil.tz import tzlocal
from persistent.TimeStamp import TimeStamp
from plone.app.caching import lastmodified
from plone.testing.zca import UNIT_TESTING
from z3c.caching.interfaces import ILastModified
from zope.component import provideAdapter

import datetime
import DateTime
import os
import time
import unittest


class FauxDataManager(object):

    def setstate(self, object):
        pass

    def oldstate(self, obj, tid):
        pass

    def register(self, object):
        pass


class TestLastModified(unittest.TestCase):

    layer = UNIT_TESTING

    def setUp(self):
        provideAdapter(lastmodified.PageTemplateDelegateLastModified)
        provideAdapter(lastmodified.FSPageTemplateDelegateLastModified)
        provideAdapter(lastmodified.OFSFileLastModified)
        provideAdapter(lastmodified.FSObjectLastModified)
        provideAdapter(lastmodified.CatalogableDublinCoreLastModified)
        provideAdapter(lastmodified.DCTimesLastModified)
        provideAdapter(lastmodified.ResourceLastModified)

    def test_PageTemplateDelegateLastModified(self):
        from persistent import Persistent
        from Acquisition import Explicit

        class Dummy(Persistent, Explicit):
            _p_mtime = None

        provideAdapter(lastmodified.PersistentLastModified, adapts=(Dummy,))

        d = Dummy()

        from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
        zpt = ZopePageTemplate('zpt').__of__(d)
        self.assertIsNone(ILastModified(zpt)())

        timestamp = 987654321.0  # time stamp (in UTC)
        # equivalent in local time, which is what the last-modified adapter
        # should return
        mod = datetime.datetime.fromtimestamp(timestamp, tzlocal())

        d._p_mtime = 987654321.0
        self.assertEqual(mod, ILastModified(zpt)())

    def test_FSPageTemplateDelegateLastModified(self):
        from persistent import Persistent
        from Acquisition import Explicit

        class Dummy(Persistent, Explicit):
            _p_mtime = None

        provideAdapter(lastmodified.PersistentLastModified, adapts=(Dummy,))

        d = Dummy()

        from Products.CMFCore.FSPageTemplate import FSPageTemplate
        zpt = FSPageTemplate('zpt', __file__).__of__(d)
        self.assertIsNone(ILastModified(zpt)())

        timestamp = 987654321.0  # time stamp (in UTC)
        # equivalent in local time, which is what the last-modified adapter
        # should return
        mod = datetime.datetime.fromtimestamp(timestamp, tzlocal())

        d._p_mtime = 987654321.0
        self.assertEqual(mod, ILastModified(zpt)())

    def test_OFSFileLastModified_File(self):
        from OFS.Image import File

        dummy = File('dummy', 'Dummy', b'data')
        self.assertIsNone(ILastModified(dummy)())

        timestamp = 987654321.0  # time stamp (in UTC)
        ts = TimeStamp(*time.gmtime(timestamp)[:6])  # corresponding TimeStamp

        # equivalent in local time, which is what the last-modified adapter
        # should return
        mod = datetime.datetime.fromtimestamp(timestamp, tzlocal())

        dummy._p_jar = FauxDataManager()
        dummy._p_serial = ts.raw()
        self.assertEqual(mod, ILastModified(dummy)())

    def test_OFSFileLastModified_Image(self):
        from OFS.Image import Image

        dummy = Image('dummy', 'Dummy', b'data')
        self.assertIsNone(ILastModified(dummy)())

        timestamp = 987654321.0  # time stamp (in UTC)
        ts = TimeStamp(*time.gmtime(timestamp)[:6])  # corresponding TimeStamp

        # equivalent in local time, which is what the last-modified adapter
        # should return
        mod = datetime.datetime.fromtimestamp(timestamp, tzlocal())

        dummy._p_jar = FauxDataManager()
        dummy._p_serial = ts.raw()
        self.assertEqual(mod, ILastModified(dummy)())

    def test_FSObjectLastModified_FSFile(self):
        from Products.CMFCore.FSFile import FSFile

        dummy = FSFile('dummy', __file__)

        modtime = float(os.path.getmtime(__file__))
        mod = datetime.datetime.fromtimestamp(modtime, tzlocal())

        # see note in test_FSObjectLastModified_FSImage
        format = '%y%m%d%H%M%s'
        self.assertEqual(mod.strftime(format),
                         ILastModified(dummy)().strftime(format))

    def test_FSObjectLastModified_FSImage(self):
        from Products.CMFCore.FSImage import FSImage

        dummy = FSImage('dummy', __file__)  # not really an image, but anyway
        modtime = float(os.path.getmtime(__file__))
        mod = datetime.datetime.fromtimestamp(modtime, tzlocal())
        # different filesystems seem to handle datetime differently.
        # Some use microseconds and others don't so to make jenkins happy,
        # lets omit the microseconds factor
        format = '%y%m%d%H%M%s'
        self.assertEqual(mod.strftime(format),
                         ILastModified(dummy)().strftime(format))

    def test_CatalogableDublinCoreLastModified(self):
        from Products.CMFCore.interfaces import ICatalogableDublinCore
        from zope.interface import implementer

        @implementer(ICatalogableDublinCore)
        class Dummy(object):

            _mod = None

            def modified(self):
                if self._mod is not None:
                    return DateTime.DateTime(self._mod)
                return None

        d = Dummy()

        self.assertIsNone(ILastModified(d)())

        d._mod = datetime.datetime(2001, 4, 19, 12, 25, 21, 120000)
        self.assertEqual(d._mod, ILastModified(d)())

    def test_DCTimesLastModified(self):
        try:
            from zope.dublincore.interfaces import IDCTimes
        except ImportError:
            return
        from zope.interface import implementer

        @implementer(IDCTimes)
        class Dummy(object):

            _mod = None

            @property
            def modified(self):
                return self._mod

        d = Dummy()

        self.assertIsNone(ILastModified(d)())

        d._mod = datetime.datetime(2001, 4, 19, 12, 25, 21, 120000)
        self.assertEqual(d._mod, ILastModified(d)())

    def test_ResourceLastModified_zope_app(self):
        from zope.browserresource.file import FileResource
        from zope.browserresource.file import File

        class DummyRequest(dict):
            pass

        request = DummyRequest()

        f = File(__file__, 'test_lastmodified.py')
        r = FileResource(f, request)

        modtime = float(os.path.getmtime(__file__))
        mod = datetime.datetime.fromtimestamp(modtime, tz=tzlocal())

        self.assertEqual(mod, ILastModified(r)())

    def test_ResourceLastModified_Five(self):
        from Products.Five.browser.resource import FileResource
        from zope.browserresource.file import File

        class DummyRequest(dict):
            pass

        request = DummyRequest()

        f = File(__file__, 'test_lastmodified.py')  # not really an image
        r = FileResource(f, request)

        modtime = float(os.path.getmtime(__file__))
        mod = datetime.datetime.fromtimestamp(modtime, tz=tzlocal())

        self.assertEqual(mod, ILastModified(r)())



# -*- coding: utf-8 -*-
from Acquisition import aq_base
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.utils import getObjectDefaultView
from plone.caching.interfaces import IRulesetLookup
from plone.registry.interfaces import IRegistry
from z3c.caching.registry import lookup
from zope.component import queryUtility
from zope.interface import implementer


@implementer(IRulesetLookup)
class ContentItemLookup(object):
    """General lookup for browser views and page templates.

    1. Attempt to look up a ruleset using z3c.caching.registry.lookup()
       and return that if found (this is necessary because this adapter will
       override the default lookup in most cases).

    2. Get the name of the published object (i.e. the name of the view or
       page template).

    3. Otherwise, look up the published name in the page template mapping (as
       PageTemplateLookup does now) and return that if found

    4. Find the parent of the published object, possibly a content object.

       4.1. If the parent is a content object:
       4.1.1. Get the default view of the parent content object
       4.1.2. If the name of the published object is the same as the default
              view of the parent:
       4.1.2.1. Otherwise, look up the parent type in the content type mapping
                and return that if found
       4.1.2.2. Look up a ruleset on the parent object and return if that
                matches

    The template mapping is:

    ``plone.app.caching.interfaces.IPloneCacheSettings.templateRulesetMapping``

    The content type mapping is:

    ``plone.app.caching.interfaces.IPloneCacheSettings.contentTypeRulesetMapping``.

    Note that this lookup is *not* invoked for a view which happens to use a
    page template to render itself.
    """

    # This adapter is registered twice in configure.zcml, ala:
    # @adapter(IPageTemplate, Interface)
    # @adapter(IBrowserView, Interface)

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):

        # 1. Attempt to look up a ruleset using the default lookup
        ruleset = lookup(self.published)
        if ruleset is not None:
            return ruleset

        registry = queryUtility(IRegistry)
        if registry is None:
            return None

        ploneCacheSettings = registry.forInterface(
            IPloneCacheSettings, check=False)

        # 2. Get the name of the published object
        name = getattr(self.published, '__name__', None)
        if name is None:
            return None

        # 3. Look up the published name in the page template mapping
        if ploneCacheSettings.templateRulesetMapping is not None:
            ruleset = ploneCacheSettings.templateRulesetMapping.get(name, None)
            if ruleset is not None:
                return ruleset

        # 4. Find the parent of the published object
        parent = getattr(self.published, '__parent__', None)
        if parent is None:
            return None

        # 4.1. If the parent is a content object:
        parentPortalType = getattr(aq_base(parent), 'portal_type', None)
        if parentPortalType is None:
            return None

        # 4.1.1. Get the default view of the parent content object
        defaultView = getObjectDefaultView(parent)

        # 4.1.2. If the name of the published object is the same as the
        # default view of the parent:
        if defaultView != name:
            return None

        # 4.1.2.1. Look up the parent type in the content type
        # mapping
        if ploneCacheSettings.contentTypeRulesetMapping is not None:
            ruleset = ploneCacheSettings.contentTypeRulesetMapping.get(
                parentPortalType,
                None,
            )
            if ruleset is not None:
                return ruleset

        # 4.1.2.2. Look up a ruleset on the parent object and
        # return
        ruleset = lookup(parent)
        if ruleset is not None:
            return ruleset



# -*- coding: utf-8 -*-
from plone.app.caching.interfaces import IRAMCached
from plone.app.caching.operations.utils import storeResponseInRAMCache
from plone.transformchain.interfaces import ITransform
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


GLOBAL_KEY = 'plone.app.caching.operations.ramcache'


@implementer(ITransform)
@adapter(Interface, Interface)
class Store(object):
    """Transform chain element which actually saves the page in RAM.

    This is registered for the ``IRAMCached`` request marker, which is set by
    the ``cacheInRAM()`` helper method. Thus, the transform is only used if
    the caching operation requested it.
    """

    order = 90000

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformUnicode(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            storeResponseInRAMCache(self.request, self.request.response,
                                    result.encode(encoding))
        return None

    def transformBytes(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            storeResponseInRAMCache(self.request, self.request.response,
                                    result)
        return None

    def transformIterable(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            result = b''.join(result)
            storeResponseInRAMCache(self.request, self.request.response,
                                    result)
            # as we have iterated the iterable, we must return a new one
            return iter([result])
        return None

    def responseIsSuccess(self):
        status = self.request.response.getStatus()
        return status == 200



# -*- coding: utf-8 -*-
from Acquisition import aq_parent
from plone.app.caching.utils import getObjectDefaultView
from plone.app.caching.utils import isPurged
from plone.behavior.interfaces import IBehaviorAssignable
from plone.cachepurging.interfaces import IPurgePathRewriter
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.interfaces import IDexteritySchema
from plone.memoize.instance import memoize
from plone.namedfile.interfaces import INamedBlobFileField
from plone.namedfile.interfaces import INamedImageField
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import IDiscussionResponse
from Products.CMFCore.interfaces import IDynamicType
from Products.CMFCore.utils import getToolByName
from z3c.caching.interfaces import IPurgePaths
from z3c.caching.purge import Purge
from zope.component import adapter
from zope.component import getAdapters
from zope.component import getUtility
from zope.event import notify
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectMovedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.schema import getFieldsInOrder

import six


try:
    from plone.app.blob.interfaces import IBlobField
    from Products.Archetypes.interfaces import IBaseObject
    from Products.Archetypes.interfaces import IFileField
    from Products.Archetypes.interfaces import IImageField
    from Products.Archetypes.interfaces import ITextField
    HAVE_AT = True
except ImportError:
    HAVE_AT = False


@implementer(IPurgePaths)
@adapter(IDynamicType)
class ContentPurgePaths(object):
    """Paths to purge for content items

    Includes:

    * ${object_path}/ (e.g. for folders)
    * ${object_path}/view
    * ${object_path}/${object_default_view}

    If the object is the default view of its parent, also purge:

    * ${parent_path}
    * ${parent_path}/
    """

    def __init__(self, context):
        self.context = context

    def getRelativePaths(self):
        prefix = '/' + self.context.virtual_url_path()
        paths = [prefix + '/', prefix + '/view']

        defaultView = getObjectDefaultView(self.context)
        if defaultView:
            path = prefix + '/' + defaultView
            if path not in paths:  # it could be
                paths.append(path)

        parent = aq_parent(self.context)
        if parent is None:
            return paths

        parentDefaultView = getObjectDefaultView(parent)
        if parentDefaultView == self.context.getId():
            parentPrefix = '/' + parent.virtual_url_path()
            paths.append(parentPrefix)
            if parentPrefix == '/':
                # special handling for site root since parentPrefix
                # does not make sense in that case.
                # Additionally, empty site roots were not getting
                # purge paths
                # /VirtualHostBase/http/site.com:80/site1/VirtualHostRoot/_vh_site1/
                # was getting generated but not
                # /VirtualHostBase/http/site.com:80/site1/VirtualHostRoot/_vh_site1
                # which would translate to http://site.come/ getting
                # invalidated but not http://site.come
                paths.append('')
                paths.append('/view')
            else:
                paths.append(parentPrefix + '/')
                paths.append(parentPrefix + '/view')

        return paths

    def getAbsolutePaths(self):
        return []


@implementer(IPurgePaths)
@adapter(IDiscussionResponse)
class DiscussionItemPurgePaths(object):
    """Paths to purge for Discussion Item.

    Looks up paths for the ultimate parent.
    """

    def __init__(self, context):
        self.context = context

    def getRelativePaths(self):
        root = self._getRoot()
        if root is None:
            return

        request = getRequest()
        if request is None:
            return

        rewriter = IPurgePathRewriter(request, None)
        for name, pathProvider in getAdapters((root,), IPurgePaths):
            # add relative paths, which are rewritten
            relativePaths = pathProvider.getRelativePaths()
            if relativePaths:
                for relativePath in relativePaths:
                    if rewriter is None:
                        yield relativePath
                    else:
                        rewrittenPaths = rewriter(
                            relativePath) or []  # None -> []
                        for rewrittenPath in rewrittenPaths:
                            yield rewrittenPath

    def getAbsolutePaths(self):
        root = self._getRoot()
        if root is None:
            return

        request = getRequest()
        if request is None:
            return

        for name, pathProvider in getAdapters((root,), IPurgePaths):
            # add absoute paths, which are not
            absolutePaths = pathProvider.getAbsolutePaths()
            if absolutePaths:
                for absolutePath in absolutePaths:
                    yield absolutePath

    @memoize
    def _getRoot(self):

        plone_utils = getToolByName(self.context, 'plone_utils', None)
        if plone_utils is None:
            return None

        thread = plone_utils.getDiscussionThread(self.context)
        if not thread:
            return None

        return thread[0]


@implementer(IPurgePaths)
@adapter(IDexteritySchema)
class ScalesPurgePaths(object):
    """Paths to purge for Dexterity object fields
    """
    def __init__(self, context):
        self.context = context

    def getScales(self):
        registry = getUtility(IRegistry)
        reg_list = registry['plone.allowed_sizes']
        sizes = [i.split(' ', 1)[0] for i in reg_list]
        sizes.append('download')
        return sizes

    def getRelativePaths(self):
        prefix = '/' + self.context.virtual_url_path()

        def fieldFilter():
            portal_type = self.context.getPortalTypeName()
            fti = getUtility(IDexterityFTI, name=portal_type)
            schema = fti.lookupSchema()
            fields = getFieldsInOrder(schema)
            assignable = IBehaviorAssignable(self.context, None)
            for behavior in assignable.enumerateBehaviors():
                new_fields = getFieldsInOrder(behavior.interface)
                if len(new_fields) > 0:
                    fields = fields + new_fields

            obj_fields = []
            for key, value in fields:
                is_image = INamedImageField.providedBy(value)
                is_file = INamedBlobFileField.providedBy(value)
                if is_image or is_file:
                    obj_fields.append(value)
            return obj_fields

        for item in fieldFilter():
            field = item.getName()
            value = item.get(self.context)
            if not value:
                continue

            if INamedImageField.providedBy(item):
                for size in self.getScales():
                    yield '{0}/@@images/{1}/{2}'.format(prefix, field, size,)
                    yield '{0}/@@download/{1}'.format(prefix, field)
            else:
                filename = value.filename
                if six.PY2 and isinstance(filename, six.text_type):
                    filename = filename.encode('utf-8')
                yield '{0}/view/{1}.{2}/@@download/{3}'.format(
                    prefix, '++widget++form.widgets', field, filename)
                yield '{0}/@@download/{1}/{2}'.format(prefix, field, filename)

    def getAbsolutePaths(self):
        return []


if HAVE_AT:

    @implementer(IPurgePaths)
    @adapter(IBaseObject)
    class ObjectFieldPurgePaths(object):
        """Paths to purge for Archetypes object fields
        """

        def __init__(self, context):
            self.context = context

        def getRelativePaths(self):
            prefix = '/' + self.context.virtual_url_path()
            schema = self.context.Schema()

            def fieldFilter(field):
                return (
                    (
                        IBlobField.providedBy(field) or
                        IFileField.providedBy(field) or
                        IImageField.providedBy(field)
                    ) and
                    not ITextField.providedBy(field)
                )
            seenDownloads = False
            for field in schema.filterFields(fieldFilter):
                if not seenDownloads:
                    yield prefix + '/download'
                    yield prefix + '/at_download'
                    seenDownloads = True

                yield prefix + '/at_download/' + field.getName()

                fieldURL = '{0}/{1}'.format(prefix, field.getName(),)
                yield fieldURL

                if IImageField.providedBy(field):
                    for size in field.getAvailableSizes(self.context).keys():
                        yield '{0}_{1}'.format(fieldURL, size,)

        def getAbsolutePaths(self):
            return []

# Event redispatch for content items - we check the list of content items
# instead of the marker interface


@adapter(IContentish, IObjectModifiedEvent)
def purgeOnModified(object, event):
    if isPurged(object):
        notify(Purge(object))


@adapter(IContentish, IObjectMovedEvent)
def purgeOnMovedOrRemoved(object, event):
    request = getRequest()
    confirmed_delete = (
        request is not None
        and getattr(request, 'URL', None)
        and 'delete_confirmation' in request.URL
        and request.REQUEST_METHOD == 'POST'
        and 'form.submitted' in request.form
    )
    if not confirmed_delete and IObjectRemovedEvent.providedBy(event):
        # ignore extra delete events
        return
    # Don't purge when added
    if IObjectAddedEvent.providedBy(event):
        return
    if isPurged(object) and 'portal_factory' not in request.URL:
        notify(Purge(object))
    parent = object.getParentNode()
    if parent:
        notify(Purge(parent))



# -*- coding: utf-8 -*-
from Acquisition import aq_base
from datetime import datetime
from dateutil.tz import tzlocal
from OFS.Image import File
from plone.app.caching.operations.utils import getContext
from Products.CMFCore.FSObject import FSObject
from Products.CMFCore.FSPageTemplate import FSPageTemplate
from Products.CMFCore.interfaces import ICatalogableDublinCore
from z3c.caching.interfaces import ILastModified
from zope.browserresource.interfaces import IResource
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from zope.pagetemplate.interfaces import IPageTemplate


try:
    from zope.dublincore.interfaces import IDCTimes
except ImportError:
    class IDCTimes(Interface):
        pass

try:
    from Products.Archetypes.Field import Image as ImageScale
except ImportError:
    class ImageScale(object):
        pass


@implementer(ILastModified)
@adapter(IPageTemplate)
def PageTemplateDelegateLastModified(template):
    """When looking up an ILastModified for a page template, look up an
    ILastModified for its context. May return None, in which case adaptation
    will fail.
    """
    return ILastModified(template.__parent__, None)


@implementer(ILastModified)
@adapter(FSPageTemplate)
def FSPageTemplateDelegateLastModified(template):
    """When looking up an ILastModified for a page template, look up an
    ILastModified for its context. Must register separately or the FSObject
    adapter would otherwise take precedence.
    """
    return PageTemplateDelegateLastModified(template)


@implementer(ILastModified)
class PersistentLastModified(object):
    """General ILastModified adapter for persistent objects that have a
    _p_mtime. Note that we don't register this for IPersistent, because
    that interface is mixed into too many things and may end up taking
    precedence over other adapters. Instead, this can be registered on an
    as-needed basis with ZCML.
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        context = aq_base(self.context)
        mtime = getattr(context, '_p_mtime', None)
        if mtime is not None and mtime > 0:
            return datetime.fromtimestamp(mtime, tzlocal())
        return None


@adapter(File)
class OFSFileLastModified(PersistentLastModified):
    """ILastModified adapter for OFS.Image.File
    """


@implementer(ILastModified)
@adapter(ImageScale)
class ImageScaleLastModified(object):
    """ILastModified adapter for Products.Archetypes.Field.Image
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        parent = getContext(self.context)
        if parent is not None:
            return ILastModified(parent)()
        return None


@implementer(ILastModified)
@adapter(FSObject)
class FSObjectLastModified(object):
    """ILastModified adapter for FSFile and FSImage
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        # Update from filesystem if we are in debug mode (only)
        self.context._updateFromFS()
        # we do this instead of getModTime() to avoid having to convert from
        # a DateTime
        mtime = self.context._file_mod_time
        return datetime.fromtimestamp(mtime, tzlocal())


@implementer(ILastModified)
@adapter(ICatalogableDublinCore)
class CatalogableDublinCoreLastModified(object):
    """ILastModified adapter for ICatalogableDublinCore, which includes
    most CMF, Archetypes and Dexterity content
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        modified = self.context.modified()
        if modified is None:
            return None
        return modified.asdatetime()


@implementer(ILastModified)
@adapter(IDCTimes)
class DCTimesLastModified(object):
    """ILastModified adapter for zope.dublincore IDCTimes
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        return self.context.modified


@implementer(ILastModified)
@adapter(IResource)
class ResourceLastModified(object):
    """ILastModified for Zope 3 style browser resources
    """

    def __init__(self, context):
        self.context = context

    def __call__(self):
        lmt = getattr(self.context.context, 'lmt', None)
        if lmt is not None:
            return datetime.fromtimestamp(lmt, tzlocal())



# -*- coding: utf-8 -*-
from hashlib import sha1 as sha
from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.cachepurging.interfaces import IPurger
from plone.protect.authenticator import _getKeyring
from zope.component import getUtility
from zope.component import provideUtility
from zope.configuration import xmlconfig
from zope.interface import implementer

import hmac


@implementer(IPurger)
class FauxPurger(object):

    def __init__(self):
        self.reset()

    def reset(self):
        self._sync = []
        self._async = []

    def purgeAsync(self, url, httpVerb='PURGE'):
        self._async.append(url)

    def purgeSync(self, url, httpVerb='PURGE'):
        self._sync.append(url)

    def stopThreads(self, wait=False):
        pass

    errorHeaders = ('X-Squid-Error',)
    http_1_1 = True


class PloneAppCaching(PloneSandboxLayer):

    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configurationContext):

        # Load ZCML
        import plone.app.caching
        xmlconfig.file('configure.zcml', plone.app.caching,
                       context=configurationContext)

        # Install fake purger
        self.oldPurger = getUtility(IPurger)
        provideUtility(FauxPurger(), IPurger)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.caching:default')

        portal['portal_workflow'].setDefaultChain(
            'simple_publication_workflow')

    def tearDownZope(self, app):
        # Store old purger
        provideUtility(self.oldPurger, IPurger)


PLONE_APP_CACHING_FIXTURE = PloneAppCaching()
PLONE_APP_CACHING_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_APP_CACHING_FIXTURE,),
    name='PloneAppCaching:Integration',
)
PLONE_APP_CACHING_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_APP_CACHING_FIXTURE,),
    name='PloneAppCaching:Functional',
)


def getToken(username):
    ring = _getKeyring(username)
    secret = ring.random().encode('utf8')
    return hmac.new(secret, username.encode('utf8'), sha).hexdigest()



# -*- coding: utf-8 -*-
from Acquisition import Explicit
from datetime import date
from datetime import datetime
from datetime import timedelta
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.utils import getObjectDefaultView
from plone.app.caching.utils import isPurged
from plone.registry import Registry
from plone.registry.fieldfactory import persistentFieldAdapter
from plone.registry.interfaces import IRegistry
from plone.testing.zca import UNIT_TESTING
from Products.CMFCore.interfaces import IDynamicType
from Products.CMFDynamicViewFTI.interfaces import IBrowserDefault
from zope.component import getUtility
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.interface import implementer

import pkg_resources
import pytz
import unittest


TEST_TIMEZONE = 'Europe/Vienna'
TEST_IMAGE = pkg_resources.resource_filename(
    u'plone.app.caching.tests', u'test.gif')


def stable_now():
    """Patch localized_now to allow stable results in tests.

    Note that a fixed date is not good enough:
    several tests compare this date with an Expires header,
    and this header may be set to ten years ago.
    """
    tzinfo = pytz.timezone(TEST_TIMEZONE)
    now = datetime.now() - timedelta(days=1000)
    now = tzinfo.localize(now)  # set tzinfo with correct DST offset
    return now


def normalize_etag(value):
    split_value = value.split('|')
    # The last component is expected to be the resourceRegistries ETag,
    # which is a time-based component, making it hard to test.
    last = split_value.pop()
    if str(date.today().year) in last:
        # yes, this is time based, remove it
        return '|'.join(split_value)
    # return original
    return value


def test_image():
    from plone.namedfile.file import NamedBlobImage
    with open(TEST_IMAGE, 'rb') as myfile:
        return NamedBlobImage(
            data=myfile.read(),
            filename=TEST_IMAGE,
        )


@implementer(IBrowserDefault, IDynamicType)
class DummyContent(Explicit):

    def __init__(self, portal_type='testtype', defaultView='defaultView'):
        self.portal_type = portal_type
        self._defaultView = defaultView

    def defaultView(self):
        return self._defaultView


class DummyNotContent(Explicit):
    pass


class DummyFTI(object):

    def __init__(self, portal_type, viewAction=''):
        self.id = portal_type
        self._actions = {
            'object/view': {'url': viewAction},
        }

    def getActionInfo(self, name):
        return self._actions[name]

    def queryMethodID(self, id, default=None, context=None):
        if id == '(Default)':
            return 'defaultView'
        elif id == 'view':
            return '@@defaultView'
        return default


@implementer(IDynamicType)
class DummyNotBrowserDefault(Explicit):

    def __init__(self, portal_type='testtype', viewAction=''):
        self.portal_type = portal_type
        self._viewAction = viewAction

    def getTypeInfo(self):
        return DummyFTI(self.portal_type, self._viewAction)


class TestIsPurged(unittest.TestCase):

    layer = UNIT_TESTING

    def setUp(self):
        provideAdapter(persistentFieldAdapter)

    def test_no_registry(self):
        content = DummyContent()
        self.assertFalse(isPurged(content))

    def test_no_settings(self):
        provideUtility(Registry(), IRegistry)
        content = DummyContent()
        self.assertFalse(isPurged(content))

    def test_no_portal_type(self):
        provideUtility(Registry(), IRegistry)
        registry = getUtility(IRegistry)
        registry.registerInterface(IPloneCacheSettings)

        ploneSettings = registry.forInterface(IPloneCacheSettings)
        ploneSettings.purgedContentTypes = ('testtype',)

        content = DummyNotContent()
        self.assertFalse(isPurged(content))

    def test_not_listed(self):
        provideUtility(Registry(), IRegistry)
        registry = getUtility(IRegistry)
        registry.registerInterface(IPloneCacheSettings)

        ploneSettings = registry.forInterface(IPloneCacheSettings)
        ploneSettings.purgedContentTypes = ('File', 'Image',)

        content = DummyContent()
        self.assertFalse(isPurged(content))

    def test_listed(self):
        provideUtility(Registry(), IRegistry)
        registry = getUtility(IRegistry)
        registry.registerInterface(IPloneCacheSettings)

        ploneSettings = registry.forInterface(IPloneCacheSettings)
        ploneSettings.purgedContentTypes = ('File', 'Image', 'testtype',)

        content = DummyContent()
        self.assertTrue(isPurged(content))


class TestGetObjectDefaultPath(unittest.TestCase):

    layer = UNIT_TESTING

    def test_not_content(self):
        context = DummyNotContent()
        self.assertIsNone(getObjectDefaultView(context))

    def test_browserdefault(self):
        context = DummyContent()
        self.assertEqual('defaultView', getObjectDefaultView(context))

    def test_browserviewdefault(self):
        context = DummyContent(defaultView='@@defaultView')
        self.assertEqual('defaultView', getObjectDefaultView(context))

    def test_not_IBrowserDefault_methodid(self):
        context = DummyNotBrowserDefault(
            'testtype', 'string:${object_url}/view')
        self.assertEqual('defaultView', getObjectDefaultView(context))

    def test_not_IBrowserDefault_default_method(self):
        context = DummyNotBrowserDefault('testtype', 'string:${object_url}/')
        self.assertEqual('defaultView', getObjectDefaultView(context))

    def test_not_IBrowserDefault_actiononly(self):
        context = DummyNotBrowserDefault(
            'testtype', 'string:${object_url}/defaultView')
        self.assertEqual('defaultView', getObjectDefaultView(context))



# -*- coding: utf-8 -*-
from Acquisition import aq_base
from Acquisition import Explicit
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.purge import ContentPurgePaths
from plone.app.caching.purge import DiscussionItemPurgePaths
from plone.app.caching.purge import HAVE_AT
from plone.app.caching.purge import purgeOnModified
from plone.app.caching.purge import purgeOnMovedOrRemoved
from plone.app.caching.purge import ScalesPurgePaths
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.contenttypes.behaviors.leadimage import ILeadImageBehavior
from plone.app.contenttypes.interfaces import IDocument
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_ROLES
from plone.behavior.interfaces import IBehavior
from plone.behavior.interfaces import IBehaviorAssignable
from plone.namedfile.file import NamedFile
from plone.namedfile.file import NamedImage
from plone.registry import Registry
from plone.registry.fieldfactory import persistentFieldAdapter
from plone.registry.interfaces import IRegistry
from plone.testing.zca import UNIT_TESTING
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import IDiscussionResponse
from Products.CMFDynamicViewFTI.interfaces import IBrowserDefault
from z3c.caching.interfaces import IPurgeEvent
from z3c.caching.interfaces import IPurgePaths
from zope.component import adapter
from zope.component import getUtility
from zope.component import provideAdapter
from zope.component import provideHandler
from zope.component import provideUtility
from zope.component import queryUtility
from zope.component.event import objectEventNotify
from zope.event import notify
from zope.globalrequest import getRequest
from zope.globalrequest import setRequest
from zope.interface import implementer
from zope.lifecycleevent import ObjectAddedEvent
from zope.lifecycleevent import ObjectModifiedEvent
from zope.lifecycleevent import ObjectMovedEvent
from zope.lifecycleevent import ObjectRemovedEvent

import six
import unittest

if HAVE_AT:
    from plone.app.caching.purge import ObjectFieldPurgePaths
    from Products.Archetypes import atapi
    from Products.Archetypes.Schema.factory import instanceSchemaFactory


def getData(filename):
    from os.path import dirname, join
    from plone.app.caching import tests

    filename = join(dirname(tests.__file__), filename)
    with open(filename, "rb") as myfile:
        return myfile.read()


class Handler(object):
    def __init__(self):
        self.invocations = []

    @adapter(IPurgeEvent)
    def handler(self, event):
        self.invocations.append(event)


class FauxRequest(dict):
    REQUEST_METHOD = "POST"
    URL = "http://nohost/test"
    form = ("form.submitted",)


@implementer(IContentish)
class FauxNonContent(Explicit):
    def __init__(self, name=None):
        self.__name__ = name
        self.__parent__ = None  # may be overridden by acquisition

    def getId(self):
        return self.__name__

    def virtual_url_path(self):
        parent = aq_base(self.__parent__)
        if parent is not None:
            return parent.virtual_url_path() + "/" + self.__name__
        else:
            return self.__name__

    def getPhysicalPath(self):
        return ("",)

    def getParentNode(self):
        return FauxNonContent("folder")


@implementer(IBrowserDefault)
class FauxContent(FauxNonContent):

    portal_type = "testtype"

    def defaultView(self):
        return "default-view"


@implementer(IDiscussionResponse)
class FauxDiscussable(Explicit):
    pass


class TestPurgeRedispatch(unittest.TestCase):

    layer = UNIT_TESTING

    def setUp(self):
        self.handler = Handler()
        provideHandler(self.handler.handler)
        provideHandler(objectEventNotify)
        provideHandler(purgeOnModified)
        provideHandler(purgeOnMovedOrRemoved)
        provideAdapter(persistentFieldAdapter)
        provideUtility(Registry(), IRegistry)
        registry = getUtility(IRegistry)
        registry.registerInterface(IPloneCacheSettings)
        ploneSettings = registry.forInterface(IPloneCacheSettings)
        ploneSettings.purgedContentTypes = ("testtype",)

    def test_not_purged(self):
        context = FauxNonContent("new").__of__(FauxContent())

        notify(ObjectModifiedEvent(context))
        notify(ObjectAddedEvent(context))
        notify(ObjectRemovedEvent(context))

        self.assertEqual(0, len(self.handler.invocations))

    def test_modified(self):
        context = FauxContent()

        notify(ObjectModifiedEvent(context))

        self.assertEqual(1, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_added(self):
        context = FauxContent("new").__of__(FauxContent())

        notify(ObjectAddedEvent(context, context.__parent__, "new"))

        self.assertEqual(0, len(self.handler.invocations))

    def test_moved(self):
        context = FauxContent("new").__of__(FauxContent())
        request = FauxRequest()
        setRequest(request)

        notify(
            ObjectMovedEvent(context, FauxContent(), "old", context.__parent__, "new")
        )

        self.assertEqual(2, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_renamed(self):
        context = FauxContent("new").__of__(FauxContent())

        notify(
            ObjectMovedEvent(
                context, context.__parent__, "old", context.__parent__, "new"
            )
        )

        self.assertEqual(2, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_removed(self):
        context = FauxContent("new").__of__(FauxContent())
        request = FauxRequest()
        request.URL = "http://nohost/delete_confirmation"
        setRequest(request)

        notify(ObjectRemovedEvent(context, context.__parent__, "new"))

        self.assertEqual(2, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)


class TestContentPurgePaths(unittest.TestCase):

    layer = UNIT_TESTING

    def test_no_default_view(self):
        context = FauxNonContent("foo")
        purger = ContentPurgePaths(context)

        self.assertEqual(["/foo/", "/foo/view"], list(purger.getRelativePaths()))
        self.assertEqual([], list(purger.getAbsolutePaths()))

    def test_default_view(self):
        context = FauxContent("foo")
        purger = ContentPurgePaths(context)

        self.assertEqual(
            ["/foo/", "/foo/view", "/foo/default-view"], list(purger.getRelativePaths())
        )
        self.assertEqual([], list(purger.getAbsolutePaths()))

    def test_parent_not_default_view(self):
        context = FauxContent("foo").__of__(FauxContent("bar"))
        purger = ContentPurgePaths(context)

        self.assertEqual(
            ["/bar/foo/", "/bar/foo/view", "/bar/foo/default-view"],
            list(purger.getRelativePaths()),
        )
        self.assertEqual([], list(purger.getAbsolutePaths()))

    def test_parent_default_view(self):
        context = FauxContent("default-view").__of__(FauxContent("bar"))
        purger = ContentPurgePaths(context)
        self.assertEqual(
            [
                "/bar/default-view/",
                "/bar/default-view/view",
                "/bar/default-view/default-view",
                "/bar",
                "/bar/",
                "/bar/view",
            ],
            list(purger.getRelativePaths()),
        )
        self.assertEqual([], list(purger.getAbsolutePaths()))


class TestDiscussionItemPurgePaths(unittest.TestCase):

    layer = UNIT_TESTING

    def setUp(self):
        @implementer(IPurgePaths)
        @adapter(FauxContent)
        class FauxContentPurgePaths(object):
            def __init__(self, context):
                self.context = context

            def getRelativePaths(self):
                return ["/" + self.context.virtual_url_path()]

            def getAbsolutePaths(self):
                return ["/purgeme"]

        provideAdapter(FauxContentPurgePaths, name="testpurge")

    def test_no_tool(self):
        root = FauxContent("")
        content = FauxContent("foo").__of__(root)
        discussable = FauxDiscussable().__of__(content)

        request = FauxRequest()
        setRequest(request)

        purge = DiscussionItemPurgePaths(discussable)

        self.assertEqual([], list(purge.getRelativePaths()))
        self.assertEqual([], list(purge.getAbsolutePaths()))

    def test_no_request(self):
        root = FauxContent("app")
        content = FauxContent("foo").__of__(root)
        discussable = FauxDiscussable().__of__(content)

        class FauxPloneTool(object):
            def getDiscussionThread(self, item):
                return [content, item]

        root.plone_utils = FauxPloneTool()

        setRequest(None)

        purge = DiscussionItemPurgePaths(discussable)

        self.assertEqual([], list(purge.getRelativePaths()))
        self.assertEqual([], list(purge.getAbsolutePaths()))

    def test_no_discussion_thread(self):
        root = FauxContent("app")
        content = FauxContent("foo").__of__(root)
        discussable = FauxDiscussable().__of__(content)

        class FauxPloneTool(object):
            def getDiscussionThread(self, item):
                return []

        root.plone_utils = FauxPloneTool()

        request = FauxRequest()
        setRequest(request)

        purge = DiscussionItemPurgePaths(discussable)

        self.assertEqual([], list(purge.getRelativePaths()))
        self.assertEqual([], list(purge.getAbsolutePaths()))

    def test_paths_of_root(self):
        root = FauxContent("app")
        content = FauxContent("foo").__of__(root)
        discussable = FauxDiscussable().__of__(content)

        class FauxPloneTool(object):
            def getDiscussionThread(self, item):
                return [content, item]

        root.plone_utils = FauxPloneTool()

        request = FauxRequest()
        setRequest(request)

        purge = DiscussionItemPurgePaths(discussable)

        self.assertEqual(["/app/foo"], list(purge.getRelativePaths()))
        self.assertEqual(["/purgeme"], list(purge.getAbsolutePaths()))


@unittest.skipUnless(HAVE_AT, "Only run with AT")
class TestObjectFieldPurgePaths(unittest.TestCase):

    maxDiff = None
    layer = UNIT_TESTING

    def setUp(self):
        provideAdapter(instanceSchemaFactory)

    def test_no_file_image_fields(self):
        class ATNoFields(atapi.BaseContent):
            schema = atapi.Schema((atapi.StringField("foo"),))

        context = ATNoFields("foo")
        purger = ObjectFieldPurgePaths(context)

        self.assertEqual([], list(purger.getRelativePaths()))
        self.assertEqual([], list(purger.getAbsolutePaths()))

    def test_file_image_fields(self):
        from plone.app.blob.field import BlobField

        class ATMultipleFields(atapi.BaseContent):
            schema = atapi.Schema(
                (
                    atapi.StringField("foo"),
                    atapi.FileField("file1"),
                    atapi.ImageField("image1"),
                    atapi.ImageField(
                        "image2", sizes={"mini": (50, 50), "normal": (100, 100)}
                    ),
                    BlobField("blob1"),
                )
            )

        root = FauxContent("")
        context = ATMultipleFields("foo").__of__(root)
        purger = ObjectFieldPurgePaths(context)

        self.assertEqual(
            [
                "/foo/download",
                "/foo/at_download",
                "/foo/at_download/file1",
                "/foo/file1",
                "/foo/at_download/image1",
                "/foo/image1",
                "/foo/image1_thumb",
                "/foo/at_download/image2",
                "/foo/image2",
                "/foo/image2_mini",
                "/foo/image2_normal",
                "/foo/at_download/blob1",
                "/foo/blob1",
            ],
            list(purger.getRelativePaths()),
        )
        self.assertEqual([], list(purger.getAbsolutePaths()))

    def test_file_image_text_fields(self):
        class ATMultipleFields(atapi.BaseContent):
            schema = atapi.Schema(
                (
                    atapi.StringField("foo"),
                    atapi.FileField("file1"),
                    atapi.ImageField("image1"),
                    atapi.ImageField(
                        "image2", sizes={"mini": (50, 50), "normal": (100, 100)}
                    ),
                    atapi.TextField("text"),
                )
            )

        root = FauxContent("")
        context = ATMultipleFields("foo").__of__(root)
        purger = ObjectFieldPurgePaths(context)

        self.assertEqual(
            [
                "/foo/download",
                "/foo/at_download",
                "/foo/at_download/file1",
                "/foo/file1",
                "/foo/at_download/image1",
                "/foo/image1",
                "/foo/image1_thumb",
                "/foo/at_download/image2",
                "/foo/image2",
                "/foo/image2_mini",
                "/foo/image2_normal",
            ],
            list(purger.getRelativePaths()),
        )
        self.assertEqual([], list(purger.getAbsolutePaths()))


class TestScalesPurgePaths(unittest.TestCase):

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):

        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Folder", "media")
        self.folder = self.portal.media
        self.folder.invokeFactory("Image", "image", title="Test Image")
        self.image_type = self.folder["image"]
        self.image_type.image = NamedImage(
            getData("data/plone-app-caching.jpg"), "image/jpg", u"plone-app-caching.jpg"
        )
        self.folder.invokeFactory("File", "file", title=u"Töst File")
        self.file = self.folder["file"]
        self.file.file = NamedFile(
            getData("data/testfile.csv"), "text/csv", u"data/töstfile.csv"
        )

        # Create a page with a lead image.
        # For the purposes of testing, we will use the Document type and
        # a custom IBehaviorAssignable adapter to mark the behavior as enabled.

        @implementer(IBehaviorAssignable)
        @adapter(IDocument)
        class TestingAssignable(object):

            enabled = [ILeadImageBehavior]
            name = "plone.leadimage"

            def __init__(self, context):
                self.context = context

            def supports(self, behavior_interface):
                return behavior_interface in self.enabled

            def enumerateBehaviors(self):
                behavior = queryUtility(IBehavior, name=self.name)
                if behavior is not None:
                    yield behavior

        provideAdapter(TestingAssignable)

        self.folder.invokeFactory("Document", "page", title="Test Page")
        self.page = self.folder["page"]

        leadimage_adapter = ILeadImageBehavior(self.page)
        leadimage_adapter.image = NamedImage(
            getData("data/plone-app-caching.jpg"), "image/jpg", u"plone-app-caching.jpg"
        )

        setRoles(self.portal, TEST_USER_ID, TEST_USER_ROLES)

    def test_scale_purge_paths(self):
        prefix = "/".join(self.image_type.getPhysicalPath())
        purge = ScalesPurgePaths(self.image_type)
        paths = purge.getRelativePaths()
        scales = purge.getScales()
        scalepaths = [prefix + "/@@images/image/" + str(i) for i in scales]
        [self.assertIn(j, paths) for j in scalepaths]
        # lead image scales (example for an image field of a behavior)
        prefix = "/".join(self.page.getPhysicalPath())
        purge = ScalesPurgePaths(self.page)
        paths = purge.getRelativePaths()
        scales = purge.getScales()
        scalepaths = [prefix + "/@@images/image/" + str(i) for i in scales]
        [self.assertIn(j, paths) for j in scalepaths]

    def test_scale_purge_paths_unicode(self):
        purge = ScalesPurgePaths(self.file)
        expected = [
            u"/plone/media/file/view/++widget++form.widgets.file/@@download/data/töstfile.csv",  # noqa: E501
            u"/plone/media/file/@@download/file/data/töstfile.csv",
        ]
        if six.PY2:
            # the getRelativePaths method returns bytes on Python 2
            expected = [x.encode("utf8") for x in expected]
        self.assertListEqual(
            list(purge.getRelativePaths()),
            expected,
        )



# -*- coding: utf-8 -*-
from plone.app.caching.interfaces import IETagValue
from plone.app.caching.operations.utils import getContext
from plone.app.caching.operations.utils import getLastModifiedAnnotation
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.resources.browser.combine import get_override_directory
from Products.CMFPlone.resources.browser.combine import PRODUCTION_RESOURCE_DIRECTORY
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.interface import Interface

import random
import time


@implementer(IETagValue)
@adapter(Interface, Interface)
class UserID(object):
    """The ``userid`` etag component, returning the current user's id
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        portal_state = queryMultiAdapter(
            (context, self.request), name=u'plone_portal_state')
        if portal_state is None:
            return None

        member = portal_state.member()
        if member is None:
            return None

        return member.getId()


@implementer(IETagValue)
@adapter(Interface, Interface)
class Roles(object):
    """The ``roles`` etag component, returning the current user's roles,
    separated by semicolons
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        portal_state = queryMultiAdapter(
            (context, self.request), name=u'plone_portal_state')
        if portal_state is None:
            return None

        if portal_state.anonymous():
            return 'Anonymous'

        member = portal_state.member()
        if member is None:
            return None

        return ';'.join(sorted(member.getRolesInContext(context)))


@implementer(IETagValue)
@adapter(Interface, Interface)
class Language(object):
    """The ``language`` etag component, returning the value of the
    HTTP_ACCEPT_LANGUAGE request key.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        return self.request.get('HTTP_ACCEPT_LANGUAGE', '')


@implementer(IETagValue)
@adapter(Interface, Interface)
class UserLanguage(object):
    """The ``userLanguage`` etag component, returning the user's preferred
    language
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        portal_state = queryMultiAdapter(
            (context, self.request), name=u'plone_portal_state')
        if portal_state is None:
            return None

        return portal_state.language()


@implementer(IETagValue)
@adapter(Interface, Interface)
class LastModified(object):
    """The ``lastModified`` etag component, returning the last modification
    timestamp
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        lastModified = getLastModifiedAnnotation(self.published, self.request)
        if lastModified is None:
            return None
        return str(time.mktime(lastModified.utctimetuple()))


@implementer(IETagValue)
@adapter(Interface, Interface)
class CatalogCounter(object):
    """The ``catalogCounter`` etag component, returning a counter which is
    incremented each time the catalog is updated.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        tools = queryMultiAdapter((context, self.request), name=u'plone_tools')
        if tools is None:
            return None

        return str(tools.catalog().getCounter())


@implementer(IETagValue)
@adapter(Interface, Interface)
class ObjectLocked(object):
    """The ``locked`` etag component, returning 1 or 0 depending on whether
    the object is locked.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        context_state = queryMultiAdapter(
            (context, self.request), name=u'plone_context_state')
        if context_state is None:
            return None
        return str(int(context_state.is_locked()))


@implementer(IETagValue)
@adapter(Interface, Interface)
class Skin(object):
    """The ``skin`` etag component, returning the current skin name.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)

        portal_skins = getToolByName(context, 'portal_skins', None)
        if portal_skins is None:
            return None

        requestVariable = portal_skins.getRequestVarname()
        if requestVariable in self.request:
            return self.request[requestVariable]

        return portal_skins.getDefaultSkin()


@implementer(IETagValue)
@adapter(Interface, Interface)
class AnonymousOrRandom(object):
    """The ``anonymousOrRandom`` etag component. This is normally added
    implicitly by the ``anonOnly`` setting. It will return None for anonymous
    users, but a random number for logged-in ones. The idea is to force a
    re-fetch of a page every time for logged-in users.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        portal_state = queryMultiAdapter(
            (context, self.request), name=u'plone_portal_state')
        if portal_state is None:
            return None
        if portal_state.anonymous():
            return None
        return '{0}{1}'.format(time.time(), random.randint(0, 1000))


@implementer(IETagValue)
@adapter(Interface, Interface)
class CopyCookie(object):
    """The ``copy`` etag component, returning 1 or 0 depending on whether
    the '__cp' cookie is set.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        return self.request.get('__cp') and '1' or '0'


@implementer(IETagValue)
@adapter(Interface, Interface)
class ResourceRegistries(object):
    """The ``resourceRegistries`` etag component, returning a timestamp.

    This is the last modified timestamp from the Plone 5+ Resource Registries.
    This is useful for avoiding requests for expired resources from cached pages.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        container = get_override_directory(context)
        if PRODUCTION_RESOURCE_DIRECTORY not in container:
            return ''
        production_folder = container[PRODUCTION_RESOURCE_DIRECTORY]
        filename = 'timestamp.txt'
        if filename not in production_folder:
            return ''
        timestamp = production_folder.readFile(filename)
        if not timestamp:
            return ''
        # timestamp is in bytes, and we must return a string.
        # On Python 2 this is the same, but not on Python 3.
        if not isinstance(timestamp, str):
            timestamp = timestamp.decode("utf-8")
        return timestamp
